
#include "Observer.h"
#include "Coordinates.h"

/// Class Coordinates
Coordenates::Coordenates(int hour, int min, float sec) {
    this->hour = hour;
    this->min = min;
    this->sec = sec;
    if (hour >= 0) {
        this->c = hour + (float)min/60 + sec/(60*60);
    } else {
        this->c = hour - (float)min/60 - sec/(60*60);
    }
}

Coordenates::Coordenates(float c) {
    this->c = c;
    // Convert a compound number into a separate Time
    this->hour = floor(abs(c));
    float m = ((abs(c) - hour) * 60);
    this->min = floor(m);
    this->sec = (m -min) * 60;

    if (c < 0) {
        this->hour *= -1;
    }
}

int Coordenates::getHour() {
    return this->hour;
}

int Coordenates::getMinutes() {
    return this->min;
}

float Coordenates::getSeconds() {
    return this->sec;
}

float Coordenates::getC() {
    return this->c;
}


/// Class Time
Time::Time(int hour, int min, float sec) : Coordenates(hour, min, sec) {
    if (hour < 0 || hour > 23 || min < 0 || min > 60 || sec < 0 || sec > 60) {
        throw new BadFormat;
    }
}

Time::Time(float c) : Coordenates(c) {
    if (c < 0 || c > 24) {
        throw new BadFormat;
    }
}

float Time::getCInDegrees() {
    return this->c * (360/24); // 15 degrees = 1 hour
}

float Time::getCInRadians() {
    return this->c * (360/24) * (PI/180);
}

/// Class Degrees
float Degrees::getCInTimeUnits() {
    return this->c / 15; // 15 degrees = 1 hour
}

float Degrees::getCInDegrees() {
    return this->c;
}

float Degrees::getCInRadians() {
    return this->c * (PI/180);
}

/// Class Degrees360
Degrees360::Degrees360(int hour, int min, float sec) : Degrees(hour, min, sec) {
    if (hour < 0 || hour > 360 || min < 0 || min > 59 || sec < 0 || sec > 60) {
        throw new BadFormat;
    }
}

Degrees360::Degrees360(float d) : Degrees(d) {
    if (d < 0 || d > 360) {
        throw new BadFormat;
    }
}

/// Class Degrees90
Degrees90::Degrees90(int hour, int min, float sec, bool negativeZero) : Degrees(hour, min, sec) {
    if (hour < -90 || hour > 90 || min < 0 || min > 59 || sec < 0 || sec > 60) {
        throw new BadFormat;
    }
    this->negativeZero = negativeZero;
    if (negativeZero) {
        this->c *= -1;
    }
}

Degrees90::Degrees90(float d) : Degrees(d) {
    if (d < -90 || d > 90) {
        throw new BadFormat;
    }
    if (d<0 && d>-1) {
        this->negativeZero = true;
    } else {
        this->negativeZero = false;
    }
}

bool Degrees90::getNegativeZero() {
    return this->negativeZero;
}

/// Class Date
Date::Date(int day, int month, int year) {
    if (day < 1 || day > 31 || month < 1 || month > 12) {
        throw new BadFormat;
    }
    this->day = day;
    this->month = month;
    this->year = year;
}

int Date::getDay() {
    return day;
}

int Date::getMonth() {
    return month;
}

int Date::getYear() {
    return year;
}
