#include "Observer.h"


Observer::Observer(Degrees90 *latitude, Time *longitude, Time *time, Time *siderealTime, Date *date) {
    this->latitude = latitude;
    this->longitude = longitude;
    this->time = time;
    this->siderealTime = siderealTime;
    this->calculateLST();
    this->date = date;

    // DEBUG
    cout << "Longitude: " << longitude->getC() << " :: " << longitude->getHour() << "h " << longitude->getMinutes() << "m " << longitude->getSeconds() << "s" <<endl;
    cout << "Latitude: " << latitude->getC() << " :: " << latitude->getHour() << "º " << latitude->getMinutes() << "' " << latitude->getSeconds() << "\"" <<endl;
    cout << "Date: " << date->getDay() << "/" << date->getMonth() << "/" << date->getYear() << endl;
    cout << "Time: " << time->getC() << " :: " << time->getHour() << "h " << time->getMinutes() << "m " << time->getSeconds() << "s" <<endl;
    cout << "ST: " << siderealTime->getC() << " :: " << siderealTime->getHour() << "h " << siderealTime->getMinutes() << "m " << siderealTime->getSeconds() << "s" <<endl;
    cout << "LST: " << localSiderealTime->getC() << " :: " << localSiderealTime->getHour() << "h " << localSiderealTime->getMinutes() << "m " <<localSiderealTime->getSeconds() << "s" <<endl;
}

void Observer::calculateLST() {
    float LST = siderealTime->getC() + longitude->getC() + 1.00273790935 * time->getC();
    if (LST >= 24) {
        LST -= 24;
    }
    this->localSiderealTime = new Time(LST);
}

// This constructor calculates siderial time from a given date
Observer::Observer(Degrees90 *latitude, Time *longitude, Time *time, Date *date) {
    this->latitude = latitude;
    this->longitude = longitude;
    this->time = time;
    this->date = date;

    // Calculate siderial time from date
    this->calculateST();

    // Calculate local siderial time
    this->calculateLST();

    // DEBUG
    cout << "Longitude: " << longitude->getC() << " :: " << longitude->getHour() << "h " << longitude->getMinutes() << "m " << longitude->getSeconds() << "s" <<endl;
    cout << "Latitude: " << latitude->getC() << " :: " << latitude->getHour() << "º " << latitude->getMinutes() << "' " << latitude->getSeconds() << "\"" <<endl;
    cout << "Date: " << date->getDay() << "/" << date->getMonth() << "/" << date->getYear() << endl;
    cout << "Time: " << time->getC() << " :: " << time->getHour() << "h " << time->getMinutes() << "m " << time->getSeconds() << "s" <<endl;
    cout << "ST: " << siderealTime->getC() << " :: " << siderealTime->getHour() << "h " << siderealTime->getMinutes() << "m " << siderealTime->getSeconds() << "s" <<endl;
    cout << "LST: " << localSiderealTime->getC() << " :: " << localSiderealTime->getHour() << "h " << localSiderealTime->getMinutes() << "m " <<localSiderealTime->getSeconds() << "s" <<endl;
}

void Observer::calculateST() {

    // Check if date is established
    if (this->date == NULL) {
        throw 1;
    }

    /// Calculate siderial time from date
    // Algorithm from http://es.wikipedia.org/wiki/Conversi%C3%B3n_de_coordenadas_ecuatoriales_a_coordenadas_horizontales
    // Julian date
    float jday = date->getDay();
    int jyear = date->getYear();
    int jmonth = date->getMonth();

    jday = jday + time->getC() / 24;

    if (date->getMonth() < 3) {
        jyear = jyear - 1;
        jmonth = jmonth + 12;
    }

    float GR = 0;
    if (jyear + jmonth / 100 + jday / 10000 >= 1582.1015)  {
        GR = 2 - floor(jyear/100) + floor(floor(jyear/100)/4);
    }

    // Julian date: number of days from 1st January 4713 b.C. at 12:00 midday
	float jdate = floor(365.25* jyear) + floor(30.6001*(jmonth+1)) + jday+1720994.5 + GR;
	float T = (jdate - 2415020)/36525;

	// Siderial time at 0h Greenwich
	float SS = 6.6460656 + 2400.051*T +0.00002581*T*T;
	float ST = (SS/24 - floor(SS/24))*24;

	this->siderealTime = new Time(ST);
}

// Only works if date has been introduced by the input file
void Observer::resetTime(Time *time){
    this->time = time;
    this->calculateST();
    this->calculateLST();
}

void Observer::resetTime(Time *time, Time *siderealTime){
    this->time = time;
    this->siderealTime = siderealTime;
    this->calculateLST();
}

Time *Observer::getSiderealTime() {
    return this->siderealTime;
}

Time *Observer::getLocalSiderealTime() {
    return this->localSiderealTime;
}

Time *Observer::getLongitude() {
    return this->longitude;
}

Degrees90 *Observer::getLatitude() {
    return this->latitude;
}






