#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>
#include <iomanip>

#include "Observer.h"
#include "CelestialObject.h"

using namespace std;


// Split a string into a vector
vector<string> &split(const string &s, char delim, vector<string> &elems) {
    stringstream ss(s);
    string item;
    while(getline(ss, item, delim)) {
        elems.push_back(item);
    }
    return elems;
}


vector<string> split(const string &s, char delim) {
    vector<string> elems;
    return split(s, delim, elems);
}


// Convert string into an integer
int toInteger(string str) {
    int i = 0;
    istringstream myStream(str);

    if (myStream>>i) {
        return i;
    } else {
        return 0;
    }
}

// Convert string into an integer
float toFloat(string str) {
    float f = 0;
    istringstream myStream(str);

    if (myStream>>f) {
        return f;
    } else {
        return 0;
    }
}

int main() {
    // Validation: http://koti.mbnet.fi/jukaukor/star_altitude.html

    // Input
    Degrees90 *latitude = NULL;
    Time *longitud = NULL;
    Date *date = NULL;
    Time *time = NULL;
    // Sidereal time: http://www.serviastro.am.ub.es/eines_calcul/calcul_dia_julia.html
    Time *siderealTime = NULL;

    Observer *observer = NULL;
    vector<CelestialObject*> objects;


    /// Read input file
    ifstream inData ("input/data.csv");
    string line;
    if (inData.is_open()) {
        while (! inData.eof() ) {
            getline (inData,line);

            // Ignore comments and empty lines
            if (line.compare(0, 1, "#") == 0 || line.empty()) {
                continue;
            }
            // Split columns
            vector<string> data = split(line, ';');

            // Get observer data
            if (data[0].compare("Latitude") == 0) {
                latitude = new Degrees90(toInteger(data[1]), toInteger(data[2]), toFloat(data[3]));
            } else if (data[0].compare("Longitud") == 0) {
                longitud = new Time(toInteger(data[1]), toInteger(data[2]), toFloat(data[3]));
            } else if (data[0].compare("Date") == 0) {
                date = new Date(toInteger(data[1]), toInteger(data[2]), toInteger(data[3]));
            } else if (data[0].compare("UniversalTime") == 0) {
                time = new Time(toInteger(data[1]), toInteger(data[2]), toFloat(data[3]));
            } else if (data[0].compare("SiderealTime") == 0) {
                siderealTime = new Time(toInteger(data[1]), toInteger(data[2]), toFloat(data[3]));
            }

            // Get celestial object data (rest of the file data)
            if (observer != NULL) {
                // DEBUG
                cout << endl << data[0] << endl;

                Time *rightAscension = new Time(toInteger(data[1]), toInteger(data[2]), toFloat(data[3]));
                Degrees90 *declination = NULL;
                // Control negative zero on degrees that cannot be represented as an in Integer in C++
                if (data[4].compare("-0")) {
                    declination = new Degrees90(toInteger(data[4]), toInteger(data[5]), toFloat(data[6]));
                } else {
                    declination = new Degrees90(toInteger(data[4]), toInteger(data[5]), toFloat(data[6]), true);
                }
                CelestialObject *obj = new CelestialObject(data[0], rightAscension, declination, data[7], observer);
                objects.push_back(obj);
            }

            // DEBUG
            //cout << data[0] << endl;

            // If we've got everything we need from the observer point of view...
            if (observer == NULL && latitude != NULL && longitud != NULL && time != NULL && date != NULL && siderealTime != NULL) {
                observer = new Observer(latitude, longitud, time, siderealTime, date);
            }

            data.clear();
        }
        inData.close();
    }


    /// Write output file
    // Output 1:
    ofstream outData;
    outData.open ("output/data.csv");
    outData << ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;" << endl;
    outData << "\"Parameters\";\"h/g\";\"m\";\"s\";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;" << endl;
    outData << "\"Longitude\";\"" << longitud->getHour() << "\";\"" << longitud->getMinutes() << "\";\"" << longitud->getSeconds() << "\";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;" << endl;
    outData << "\"Latitude\";\"" << latitude->getHour() << "\";\"" << latitude->getMinutes() << "\";\"" << latitude->getSeconds() << "\";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;" << endl;
    outData << "\"Time\";\"" << time->getHour() << "\";\"" << time->getMinutes() << "\";\"" << time->getSeconds() << "\";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;" << endl;
    outData << "\"Sidereal Time\";\"" << siderealTime->getHour() << "\";\"" << siderealTime->getMinutes() << "\";\"" << siderealTime->getSeconds() << "\";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;" << endl;
    outData << "\"Date\";\"" << date->getDay() << "\";\"" << date->getMonth() << "\";\"" << date->getYear() << "\";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;" << endl;
    outData << ";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;" << endl;
    outData << "\"Name\";\"Right Ascension\";;;\"Declination\";;;\"Azimuth (South = 0)\";;;\"Altitude\";;;\"Culmination\";;;\"Magnitude\"" << endl;
    outData << ";\"h\";\"m\";\"s\";\"g\";\"m\";\"s\";\"g\";\"m\";\"s\";\"g\";\"m\";\"s\";\"h\";\"m\";\"s\";" << endl;

    // Output 2: Positions in horizontal coordinates
    ofstream outGnuplot;
    outGnuplot.open ("output/position/objects.dat");

    ofstream gnuplotPositionScript;
    string filename = "output/position/visualizeHorizontal.plot";
    gnuplotPositionScript.open (filename.c_str());
    gnuplotPositionScript << "#!/usr/bin/gnuplot" << endl;
    gnuplotPositionScript << "unset key" << endl;
    gnuplotPositionScript << "set title 'Object positions'" << endl;
    gnuplotPositionScript << "set yrange [0:90]" << endl;
    gnuplotPositionScript << "set xrange [0:360]" << endl;
    gnuplotPositionScript << "set xtics 45" << endl;
    gnuplotPositionScript << "set ytics 10" << endl;
    gnuplotPositionScript << "set grid" << endl;
    gnuplotPositionScript << "set xlabel 'Azimuth'" << endl;
    gnuplotPositionScript << "set ylabel 'Altitude'" << endl;
    gnuplotPositionScript << "set rmargin 10" << endl;
    gnuplotPositionScript << "set lmargin 10" << endl;
    gnuplotPositionScript << "set label \"South\"  at first -15, first 93" << endl;
    gnuplotPositionScript << "set label \"North\"  at first 165, first 93" << endl;
    gnuplotPositionScript << "set label \"South\"  at first 345, first 93" << endl;
    gnuplotPositionScript << "plot \"objects.dat\" using 2:3:1 with labels font \"Helvetica,8\" offset 2,0 point ps 1 pt 3" << endl;
    gnuplotPositionScript << "pause -1" << endl;
    gnuplotPositionScript.close();

    // Output 3: Positions in cartesian coordinates
    ofstream outSouthGnuplot;
    outSouthGnuplot.open ("output/position/southObjects.dat");
    ofstream outNorthGnuplot;
    outNorthGnuplot.open ("output/position/northObjects.dat");

    ofstream gnuplotCartesianPositionScript;
    filename = "output/position/visualizeCartesian.plot";
    gnuplotCartesianPositionScript.open (filename.c_str());
    gnuplotCartesianPositionScript << "#!/usr/bin/gnuplot" << endl;
    gnuplotCartesianPositionScript << "unset key" << endl;
    gnuplotCartesianPositionScript << "set title 'Object positions'" << endl;
    gnuplotCartesianPositionScript << "set zrange [0:1]" << endl;
    gnuplotCartesianPositionScript << "set view 180,270,1,1; # Looking up" << endl;
    /*#set view 90,0,1,1; # Pointing east
    #set view 90,90,1,1; # Pointing south
    #set view 90,270,1,1; # Pointing north
    #set view 90,360,1,1; # Pointing west*/
    //gnuplotMotionScript << "set palette defined (-3 'blue', 0 'white', 1 'red')"
    /*gnuplotCartesianPositionScript << "set terminal png size 800, 600" << endl;
    gnuplotCartesianPositionScript << "set output \"cartesianPosition.png\"" << endl;*/
    gnuplotCartesianPositionScript << "splot \"northObjects.dat\" using 2:3:4:1 with labels font \"Helvetica,8\" offset 2,0 point ps 1 pt 3, \"southObjects.dat\" using 2:3:4:1 with labels font \"Helvetica,8\" offset 2,0 point ps 1 pt 3 tc lt 9;" << endl;
    gnuplotCartesianPositionScript << "pause -1" << endl;
    gnuplotCartesianPositionScript.close();


    // Output 4: Generate a gnuplot script in order to create PNG files with celestial objects motion
    ofstream gnuplotMotionScript;
    filename = "output/motion/generateImages.plot";
    gnuplotMotionScript.open (filename.c_str());
    gnuplotMotionScript << "#!/usr/bin/gnuplot" << endl;
    gnuplotMotionScript << "set yrange [-90:90]" << endl;
    gnuplotMotionScript << "set xrange [0:24]" << endl;
    gnuplotMotionScript << "set xtics 1" << endl;
    gnuplotMotionScript << "set ytics 10" << endl;
    gnuplotMotionScript << "set grid" << endl;
    gnuplotMotionScript << "set xlabel 'Universal time'" << endl;
    gnuplotMotionScript << "set ylabel 'Altitude'" << endl;
    gnuplotMotionScript << "set key top left" << endl;
    gnuplotMotionScript << "set key box" << endl;
    gnuplotMotionScript << "set rmargin 10" << endl;
    gnuplotMotionScript << "set lmargin 10" << endl;


    vector<CelestialObject*>::iterator obj;
    for (obj = objects.begin(); obj != objects.end(); obj++) {
        CelestialObject *c = *obj;
        Degrees360 *a = c->getAzimuth();
        Degrees90 *h = c->getAltitude();
        Time *t = c->getTimeOfCulmination();
        Time *ra = c->getRightAscension();
        Degrees90 *d = c->getDeclination();

        // Output 1:
        // Name
        outData << "\"" << c->getName() << "\";";
        // Right Ascension
        outData << "\"" << ra->getHour() << "\";\"" << ra->getMinutes() << "\";\"" << setprecision(6) << ra->getSeconds() << "\";";
        // Declination
        if (d->getNegativeZero()) {
            // Write a minus in front of the negative zero
            outData << "\"-" << d->getHour() << "\";\"" << d->getMinutes() << "\";\"" << setprecision(6) << d->getSeconds() << "\";";
        } else {
            outData << "\"" << d->getHour() << "\";\"" << d->getMinutes() << "\";\"" << setprecision(6) << d->getSeconds() << "\";";
        }
        // Azimuth
        outData << "\"" << a->getHour() << "\";\"" << a->getMinutes() << "\";\"" << setprecision(6) << a->getSeconds() << "\";";
        // Altitude
        if (h->getNegativeZero()) {
            // Write a minus in front of the negative zero
            outData << "\"-" << h->getHour() << "\";\"" << h->getMinutes() << "\";\"" << setprecision(6) << h->getSeconds() << "\";";
        } else {
            outData << "\"" << h->getHour() << "\";\"" << h->getMinutes() << "\";\"" << setprecision(6) << h->getSeconds() << "\";";
        }
        // Culmination time
        outData << "\"" << t->getHour() << "\";\"" << t->getMinutes() << "\";\"" << setprecision(6) << t->getSeconds() << "\";";
        // Magnitude
        outData << "\"" << c->getMagnitude() << "\"" << endl;

        /// GNUPLOT
        // Output 2: Positions in horizontal coordinates
        outGnuplot << "\"" << c->getName() << "\"\t" << a->getC() << "\t" << h->getC() << endl;

        // Output 3: Convert into cartesian coordinates
        float x = cos(h->getCInRadians())*cos(a->getCInRadians());
        float y = cos(h->getCInRadians())*sin(a->getCInRadians());
        float z = sin(h->getCInRadians());

        // North or South
        if (x <= 0) {
            outNorthGnuplot << "\"" << c->getName() << "\"\t" << x << "\t" << y << "\t" << z << "\t" << a->getC() << "\t" << h->getC() << endl;
        } else {
            outSouthGnuplot << "\"" << c->getName() << "\"\t" << x << "\t" << y << "\t" << z << "\t" << a->getC() << "\t" << h->getC() << endl;
        }

        // Output 4:
        // For each celestial object simulate their motion from 0 to 24 hours
        // - Generate a .dat file with time and altitude relation
        ofstream outMotion;
        string filename = "output/motion/" + c->getName() + ".dat";
        outMotion.open (filename.c_str());

        Time *newTimeObj = new Time(0);
        for(int i=0; i < 48; i ++) {
            // Recalculate altitude in function of time changes
            observer->resetTime(newTimeObj);
            c->recalculateHorizontalCoordinates();
            outMotion << newTimeObj->getC() << "\t" << c->getAltitude()->getCInDegrees() << endl;

            // Check overflow
            float newTime = newTimeObj->getC() + 0.5;
            if (newTime >= 24.0) {
                newTime -= 24;
            }
            delete newTimeObj;
            newTimeObj = new Time(newTime);
        }
        delete newTimeObj;
        // Reset observer to normal time
        observer->resetTime(time, siderealTime);
        outMotion.close();

        // Complete gnuplot script in order to generate the PNG file with the motion
        gnuplotMotionScript << "set title \"" + c->getName() + "\"" << endl;
        gnuplotMotionScript << "set terminal png size 800, 600" << endl;
        gnuplotMotionScript << "set output \"" + c->getName() + ".png\"" << endl;

        // Mark the observartion hour
        gnuplotMotionScript << "set arrow from " << time->getC() << ",90 to " << time->getC() << ",-90 " << endl;
        gnuplotMotionScript << "set label \" Observation\\n hour\"  at first  " << time->getC() << ", first -70" << endl;

        // Mark the culmination hour
        gnuplotMotionScript << "set arrow from " << c->getTimeOfCulmination()->getC() << ",90 to " << c->getTimeOfCulmination()->getC() << ",-90 ls 8" << endl;
        gnuplotMotionScript << "set label \" Culmination\\n hour\"  at first  " << c->getTimeOfCulmination()->getC() << ", first 70 textcolor lt 8" << endl;

        string line = "plot \"" + c->getName() + ".dat\" using 1:2 with line title 'Motion', 0 title 'Skyline'";
        gnuplotMotionScript << line << endl;

        gnuplotMotionScript << "unset label" << endl;
        gnuplotMotionScript << "unset arrow" << endl;


    }

    gnuplotMotionScript.close();
    outSouthGnuplot.close();
    outNorthGnuplot.close();
    outGnuplot.close();
    outData.close();

    return 0;
}


